/**
 * \file Config.h
 *
 *  \date Jan 31, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBONCILLA_WEBOTS_CONFIG_H_
#define LIBONCILLA_WEBOTS_CONFIG_H_


#include <libatuleu/options/Option.h>
#include <libatuleu/base/Singleton.h>
/**
 *
 */

namespace liboncilla {
namespace webots {
namespace config{

	class SectionConfig {
	public :
		SectionConfig(options::Section::Ptr base);
		virtual ~SectionConfig();
		options::Section::Ptr & Section();
	private :
		options::Section::Ptr d_section;
	};

#define ADD_OPTION_IMPLEMENTATION(name, type) private :\
	options::Option<type>::Ptr d_o ## name ;\
	public :\
	const type & name () const { return d_o ## name ->Value(); }

#define ADD_BOOL_IMPLEMENTATION(name) private:\
	options::Option<bool>::Ptr d_o ## name ;\
	public : \
	bool name () const { return d_o ## name -> Value();}


	class PassiveDynamicsSection : public SectionConfig{
	public :
		void SendJointDefinition(uint32_t leg_identifier) const;

		PassiveDynamicsSection(options::Section::Ptr base);
		virtual ~PassiveDynamicsSection();

		ADD_OPTION_IMPLEMENTATION(Stiffness,double);
		ADD_OPTION_IMPLEMENTATION(PreCompression,double);
		ADD_OPTION_IMPLEMENTATION(Damping,double);
		ADD_OPTION_IMPLEMENTATION(MaxLength, double);
		ADD_OPTION_IMPLEMENTATION(MinLength, double);

	};




	class LegSection : public SectionConfig{
	public :
		LegSection(options::Section::Ptr base);
		virtual ~LegSection();

		const PassiveDynamicsSection & Diagonal() const;
		const PassiveDynamicsSection & Parallel() const;
		const PassiveDynamicsSection & Foot() const;
		ADD_OPTION_IMPLEMENTATION(DiagonalErp,double);
		ADD_OPTION_IMPLEMENTATION(DiagonalCfm,double);
		ADD_OPTION_IMPLEMENTATION(Margin,double);
		ADD_OPTION_IMPLEMENTATION(MaxSpeed,double);
	private :
		PassiveDynamicsSection d_diagonal,d_parallel,d_foot;
	};


	class ContactSection : public SectionConfig {
	public :
		ContactSection(options::Section::Ptr base);
		virtual ~ContactSection();

		ADD_OPTION_IMPLEMENTATION(Bounce,double);
		ADD_OPTION_IMPLEMENTATION(BounceVelocity,double);
		ADD_OPTION_IMPLEMENTATION(CoulombFriction,double);
		ADD_OPTION_IMPLEMENTATION(Fds,double);
		ADD_OPTION_IMPLEMENTATION(Erp,double);
		ADD_OPTION_IMPLEMENTATION(Cfm,double);
		ADD_BOOL_IMPLEMENTATION(Draw)
	};



	class MainSection : public SectionConfig{
	public :
		MainSection(options::Section::Ptr base);
		virtual ~MainSection();

		ADD_BOOL_IMPLEMENTATION(DebugLibwebots);
		ADD_OPTION_IMPLEMENTATION(InitTime,double);
		ADD_OPTION_IMPLEMENTATION(ProcessLoopTime,double);

	};
	class Config : public Singleton<Config> {
	public :
		Config();
		virtual ~Config();
		const LegSection & ForeLeg();
		const LegSection & HindLeg();
		const ContactSection & Contact();
		const MainSection & Main();

		void LoadConfigFile();
	private :
		options::Section::Ptr d_root;
		LegSection d_foreLeg;
		LegSection d_hindLeg;
		MainSection d_main;
		ContactSection d_contact;

	};

};
} /* namespace liboncilla */
} /* namespace webots */

#endif // LIBONCILLA_WEBOTS_CONFIG_H_
