/*
 * OncillaWebotsSynchronizer.h
 *
 *  Created on: 15 d�c. 2011
 *      Author: Alexandre Tuleu
 */

#ifndef LIBONCILLA_WEBOTS_ONCILLAWEBOTSSYNCHRONIZER_H_
#define LIBONCILLA_WEBOTS_ONCILLAWEBOTSSYNCHRONIZER_H_

#include <webots/Supervisor.hpp>
#include <boost/shared_ptr.hpp>
#include <memory>

#include <liboncilla/OncillaSynchronizer.h>
#include <liboncilla/BackendFactory.h>

#include <Eigen/Core>

#include <stdexcept>

#include "LegParameters.h"

#include "WebotsNodeWrapper.h"

extern "C" {
	void liboncilla_webots_backend_entry();
}

namespace liboncilla{
namespace webots{
namespace config{
	class LegSection;
}
}
}

class WebotsNodeAdapter;



class OncillaWebotsSynchronizer : public rci::oncilla::OncillaSynchronizer{

public:
	static rci::oncilla::OncillaSynchronizer * Create();


	OncillaWebotsSynchronizer(webots::Supervisor * supervisor,
	                          const std::string & name ="Oncilla");
	virtual ~OncillaWebotsSynchronizer();

	virtual void registerTrunkNode(rci::oncilla::OncillaTrunkPtr node);

	virtual void registerL0Node(rci::oncilla::Leg l,
	                            rci::oncilla::OncillaL0Ptr node);
	virtual void registerL1Node(rci::oncilla::Leg l,
	                            rci::oncilla::OncillaL1Ptr node);
	virtual void registerL2Node(rci::oncilla::Leg l,
	                            rci::oncilla::OncillaL2Ptr node);
	virtual void registerL3Node(rci::oncilla::Leg l,
	                            rci::oncilla::OncillaL3Ptr node);
	virtual void registerL4Node(rci::oncilla::Leg l,
	                            rci::oncilla::OncillaL4Ptr node);

	virtual void processAsync();
	virtual void waitForProcessAsync();

	virtual bool start();
	virtual bool stop();

	const LegParameters & GetLegParameters(rci::oncilla::Leg l);

	virtual double lastProcessTimeStep() const;

private :
	typedef std::vector<std::string>                LegNames;
	typedef std::auto_ptr<webots::Supervisor>       WSupervisorPtr;
	typedef boost::shared_ptr<WebotsNodeAdapter> WebotsNodeAdapterPtr;
	typedef std::vector<WebotsNodeAdapterPtr>       ListOfAdpater;
	typedef std::vector<LegParameters>              ListOfLegParameters;

	enum JointIdentifier {
		TRUNK_L0 = 0x10,
		L0_L1    = 0x20,
		L1_L2    = 0x30,
		L2_L3    = 0x40,
		L1_D1    = 0x50,
		D1_D2    = 0x60,
		D2_L3    = 0x70,
		L1_P1    = 0x80,
		P1_P2    = 0x90,
		P2_L3    = 0xA0,
		L3_FOOT  = 0xB0,
		P1_L3    = 0xC0
	};

	static LegNames s_webotsPrefix;

	void ExtractLegParametersAndInitModel(rci::oncilla::Leg l,
	                          LegParameters & params);
	void ExtractLegParameters(const WebotsNodeWrapper & l1,
	                          const WebotsNodeWrapper & l2,
	                          const WebotsNodeWrapper & l3,
	                          const WebotsNodeWrapper & p1,
	                          const WebotsNodeWrapper & foot,
	                          const std::string & d2name,
	                          LegParameters & params);

	const liboncilla::webots::config::LegSection & LegConfigSection(rci::oncilla::Leg l);

	uint32_t GetJointIdentifier(rci::oncilla::Leg l,
	                            JointIdentifier j);

	void Step();

	WSupervisorPtr d_supervisor;
	ListOfAdpater  d_adapters;
	std::vector<LegParameters> d_legParameters;
	bool d_initialized;
	double d_processTimeStep;
};

inline uint32_t
OncillaWebotsSynchronizer::GetJointIdentifier(rci::oncilla::Leg l ,
                                              OncillaWebotsSynchronizer::JointIdentifier j){
	return l + j;
}

inline double GetAbsMaxOfArray(const double * val){
       return std::max(std::max(std::abs(val[0]),std::abs(val[1])),std::abs(val[2]));
}




#endif /* LIBONCILLA_WEBOTS_ONCILLAWEBOTSSYNCHRONIZER_H_ */
