/*
 * WebotsL1Adapter.cpp
 *
 *  Created on: 15 déc. 2011
 *      Author: Alexandre Tuleu
 */

#include "WebotsL1Adapter.h"
#include "Config.h"
using namespace rci::oncilla;
using namespace webots;

template <typename T> const T & clamp(const T & value, const T & min ,const  T &max){
	return std::min<T>(std::max<T>(value,min),max);
}

WebotsL1Adapter::WebotsL1Adapter(OncillaL1Ptr node,
                                 Servo * servo,
                                 const WebotsNodeWrapper & wNode)
	: d_node(node)
	, d_servo(*servo)
	, d_desiredPosition(0.0)
	, d_maxServoForce(wNode.GetField<double>("maxForce"))
	, d_rate(0.0)
	, d_lastServoPosition(0.0)
	, d_maxPosition(wNode.GetField<double>("maxPosition"))
	, d_minPosition(wNode.GetField<double>("minPosition")){
		if(!servo){
		throw std::logic_error("Unvalid webots::Servo handle.");
	}


}

WebotsL1Adapter::~WebotsL1Adapter() {

}

void WebotsL1Adapter::PreStep(double ts){
	d_rate = d_servo.getPosition() - d_lastServoPosition;
	d_rate /= ts;
	d_lastServoPosition = d_servo.getPosition();

	rci::ControlModePtr c = d_node->getControlMode();
	if(c->supports(rci::jointPositionControl)){


		d_desiredPosition = clamp(d_node->getLastPositionCommand()->asDouble(0),
		                          d_minPosition,d_maxPosition);
		d_servo.setPosition(d_desiredPosition);


	}else if (c->supports(rci::jointVelocityControl)){
		d_servo.setVelocity(d_node->getLastVelocityCommand()->asDouble(0));
	} else if (c->supports(rci::jointTorqueControl)) {
		d_servo.setMotorForce(d_node->getLastTorqueCommand()->asDouble(0));
	} else {
		throw std::runtime_error("you should select either position, velocity or torque control.");
	}
	double force(d_maxServoForce);

	if(c->supports(rci::jointComplianceControl)) {
		double stiff = d_node->getLastImpedanceCommand()->getStiffness(0).asDouble(0);
		double damp = d_node->getLastImpedanceCommand()->getDamping(0).asDouble(0);
		double precomp = 0;//should be in Impedance;
		force = stiff * (d_desiredPosition - d_servo.getPosition() - precomp) - damp * d_rate;
	}

	//set the maximal force that we could use
	d_servo.setMotorForce(std::abs(force));

}


void WebotsL1Adapter::PostStep(double){

	rci::JointAnglesPtr encoder(new rci::JointAngles(nemo::RealVector(d_servo.getPosition(),
	                                                                  d_servo.getPosition())));

	d_node->unsafeUpdateMotorAxisPosition(d_servo.getPosition());
	d_node->unsafeUpdateMagneticEncoderPosition(d_servo.getPosition());

}

void WebotsL1Adapter::Activate(){
	namespace c = liboncilla::webots::config;
	d_servo.enablePosition(c::Config::Instance().Main().ProcessLoopTime() * 1000);
}

void WebotsL1Adapter::DeActivate(){
	d_servo.disablePosition();
}

