/*
 * WebotsL1Adapter.h
 *
 *  Created on: 15 d�c. 2011
 *      Author: Alexandre Tuleu
 */

#ifndef LIBONCILLA_WEBOTS_WEBOTSL1ADAPTER_H_
#define LIBONCILLA_WEBOTS_WEBOTSL1ADAPTER_H_


#include <liboncilla/OncillaL1L2.h>
#include <webots/Servo.hpp>
#include <webots/Node.hpp>

#include "WebotsNodeAdapter.h"
#include "WebotsNodeWrapper.h"


/**
 * Adapter for OncillaL1 joint.
 *
 * \section Modelisation
 *
 * This joint only uses a standard webots servo, which can already be
 * positionned/velocity/torque controlled. The virtual impedance control is
 * implemented like on the real robot, by limiting the maximal torque the PID
 * can internally use to reach a given position, using a
 * spring-damper-precompression equation.
 *
 *
 * However there is a tiny difference at this point of writing. On the real
 * robot the virtual compliance is computed every ms, independantly of the
 * frequency at wich the user send data to the robot. Here this frequency may be
 * different since it is the timestep of integration that may be higher than 1.
 * One solution will be to implement this in the physics plugin.
 */

class WebotsL1Adapter : public WebotsNodeAdapter{
public:
	WebotsL1Adapter(rci::oncilla::OncillaL1Ptr node,
	                webots::Servo * servo,
	                const WebotsNodeWrapper & wNode);
	virtual ~WebotsL1Adapter();
	void PreStep(double ts);
	void PostStep(double ts);
	void Activate();
	void DeActivate();

private :
	rci::oncilla::OncillaL1Ptr d_node;
	webots::Servo & d_servo;

	double d_desiredPosition;
	double d_maxServoForce;
	double d_rate;
	double d_lastServoPosition;
	double d_maxPosition,d_minPosition;
};

#endif /* LIBONCILLA_WEBOTS_WEBOTSL1ADAPTER_H_ */
