/**
 * \file WebotsL4Adapter.cpp
 *
 *  \date Jan 27, 2012
 *  \author Alexandre Tuleu
 */

#include "WebotsL4Adapter.h"

#include <libwebots/communication/client.hh>
#include <libwebots/communication/messages.hh>
#include "Config.h"

using namespace rci::oncilla;

WebotsL4Adapter::ContactManager WebotsL4Adapter::s_cManager;

WebotsL4Adapter::WebotsL4Adapter(OncillaL4Ptr node,
            		             const std::string & footName)
	: d_node(node){
	s_cManager.RegisterNewContact(footName,this);
	//s_cManager.RegisterNewContact(l3Name,this);
}

WebotsL4Adapter::~WebotsL4Adapter(){
}

void WebotsL4Adapter::PreStep(double ){
	s_cManager.Reset();
}

void WebotsL4Adapter::PostStep(double){
	d_node->unsafeUpdateForce(d_data.block<3,1>(0,0).data());
}
void WebotsL4Adapter::Activate(){}
void WebotsL4Adapter::DeActivate(){}


WebotsL4Adapter::ContactManager::ContactManager()
	: d_state(UNINITIALIZED){
}

void WebotsL4Adapter::ContactManager::Reset(){
	if(d_state == UNINITIALIZED){
		Init();
	}

	if(d_state == SOME_DATA_INPUTED){
		for(AdapterByName::iterator a = d_adapters.begin();
		    a != d_adapters.end();
		    ++a){
			a->second->d_data.setZero();
		}

		d_state = DATA_RESETED;
	}
}

void WebotsL4Adapter::ContactManager::Init(){
	using namespace libwebots::communication;

	if(d_state != UNINITIALIZED){
		return;
	}

	Client::Instance().AddHandler(*this,
	                              lmc::contactmonitor,
	                              &ContactManager::Handler);

	d_state = SOME_DATA_INPUTED;
}


void WebotsL4Adapter::ContactManager::Handler(lmc::Monitor const & m){
	AdapterByName::const_iterator a = d_adapters.find(m.name());
	if(a == d_adapters.end()){
		a = d_adapters.find(m.other());
	}

	if ( a == d_adapters.end()){
		return;
	}


	Eigen::Vector3d position,value;
	position.setZero();value.setZero();
	double totalNorm = 0.0;
	unsigned int nbContact = m.position_size() / 3;
	for(unsigned int i = 0 ; i < nbContact; ++i){
		Eigen::Vector3d partPosition,partValue;
		partPosition << m.position(3 * i    )
		             ,  m.position(3 * 1 + 1)
		             ,  m.position(3 * i + 2);

		partValue << m.force(3 * i    )
		          ,  m.force(3 * i + 1)
		          ,  m.force(3 * i + 2);

		value += partValue;

		double norm = partValue.norm();
		position += partPosition * norm;
		totalNorm += norm;
	}
	if(totalNorm > 0.0){
		position /= totalNorm;
	}

	a->second->d_data.block<3,1>(0,0) = value;
	a->second->d_data.block<3,1>(3,0) = position;
}

const char * WebotsL4Adapter::ContactManager::GroundBodyName = "GROUND";

void
WebotsL4Adapter::ContactManager::RegisterNewContact(const std::string & name,
                                                    WebotsL4Adapter * a){
	Init();
	using namespace liboncilla::webots::config;
	using namespace libwebots::communication;
	const ContactSection & c = Config::Instance().Contact();
	lmc::Contact m;
	m.set_body1(name);
	m.set_body2(GroundBodyName);
	m.set_coulombfriction(c.CoulombFriction());
	m.set_draw(c.Draw());
	m.set_monitor(true);

	Client::Instance().Send(Messages::Create(lmc::contact,m));

	d_adapters.insert(std::make_pair(name,a));
	d_state = SOME_DATA_INPUTED;
}

