/**
 * \file WebotsL4Adapter.h
 *
 *  \date Jan 27, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBONCILLA_WEBOTS_WEBOTSL4ADAPTER_H_
#define LIBONCILLA_WEBOTS_WEBOTSL4ADAPTER_H_

#include <liboncilla/OncillaL4.h>
#include "WebotsNodeAdapter.h"

#include <libwebots/plugin/modules/contact/contact.pb.h>

#include <Eigen/Core>

#if EIGEN_WORLD_VERSION < 3
#error "inclusion error, I need eigen 3"
#endif

namespace lmc = libwebots::plugin::modules::messages::contact;

/**
 *
 */
class WebotsL4Adapter : public WebotsNodeAdapter{
public:


	class ContactManager {
	public :
		ContactManager();
		void Handler(lmc::Monitor const & m);
		void Reset();
		void RegisterNewContact(const std::string & name,
		                        WebotsL4Adapter * adapter);
	private :
		typedef std::map<std::string,WebotsL4Adapter*> AdapterByName;
		enum State{
			UNINITIALIZED,
			DATA_RESETED,
			SOME_DATA_INPUTED
		};

		void Init();

		AdapterByName d_adapters;
		State d_state;
		static const char * GroundBodyName;
	};


	WebotsL4Adapter(rci::oncilla::OncillaL4Ptr node,
	                const std::string & footName);
	virtual ~WebotsL4Adapter();

	void PreStep(double ts);
	void PostStep(double ts);
	void Activate();
	void DeActivate();

private:
	typedef Eigen::Matrix<double,6,1> SensorData;

	static ContactManager s_cManager;

	void RegisterNewGroundContactMonitor(const std::string & bodyName);

	rci::oncilla::OncillaL4Ptr d_node;

	SensorData d_data;


};

#endif // LIBONCILLA_WEBOTS_WEBOTSL4ADAPTER_H_
