/**
 * \file WebotsNodeWrapper.h
 *
 *  \date Jun 27, 2012
 *  \author Alexandre Tuleu
 */

#ifndef LIBONCILLA_WEBOTS_WEBOTSNODEWRAPPER_H_
#define LIBONCILLA_WEBOTS_WEBOTSNODEWRAPPER_H_

#include <stdexcept>
#include <string>


#include <webots/Node.hpp>

namespace webots{
	class Supervisor;
	class Node;
}

/**
 *
 */
class WebotsNodeWrapper {
public:
	static void SetSupervisor(webots::Supervisor * supervisor);

	WebotsNodeWrapper(const std::string & name);
	virtual ~WebotsNodeWrapper();

	template <typename T>
	T GetField(const std::string & name) const;

	int Type() const;
private :
	static webots::Supervisor * s_supervisor;
	webots::Node * d_node;
};

template <>
inline double WebotsNodeWrapper::GetField<double>(const std::string & name) const {
	webots::Field * f = d_node->getField(name);
	if(f == 0 ){
		throw std::runtime_error("[WebotsNodeWrapper] : could not find field '" + name + "' in node '" + d_node->getName() + "'");
	}
	return f->getSFFloat();
}

template <>
inline const double * WebotsNodeWrapper::GetField<const double *>(const std::string & name) const {
	webots::Field * f = d_node->getField(name);
	if(f == 0 ){
		throw std::runtime_error("[WebotsNodeWrapper] : could not find field '" + name + "' in node '" + d_node->getName() + "'");
	}
	return f->getSFVec3f();
}


template <>
inline webots::Field * WebotsNodeWrapper::GetField<webots::Field *>(const std::string & name) const {
	webots::Field * f = d_node->getField(name);
	if(f == 0 ){
		throw std::runtime_error("[WebotsNodeWrapper] : could not find field '" + name + "' in node '" + d_node->getName() + "'");
	}
	return f;
}



#endif // LIBONCILLA_WEBOTS_WEBOTSNODEWRAPPER_H_
